from scipy.ndimage.filters import laplace
from skimage import measure
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
from scipy.interpolate import griddata


class tomo_laplace():
    def __init__(self, X, Z, RHO, ax):
        self.X = X
        self.Z = Z
        self.RHO = RHO
        self.dl = np.abs(X[0, 0] - X[0, 1])
        self.ax = ax

        self.lap_edge()
        self.plot_laplacian_edges()

    def lap_edge(self):
        self.lap_grid = laplace(self.RHO)  # Laplacian
        # Gradient magnitude
        sx = ndimage.sobel(self.RHO, axis=0, mode='reflect')
        sy = ndimage.sobel(self.RHO, axis=1, mode='reflect')
        self.sobel = np.hypot(sx, sy)  # Gradient magnitude

    def plot_laplacian_edges(self):
        contours = measure.find_contours(self.lap_grid, 0.0)  # contours where Laplacian=0

        x_int_all = []
        z_int_all = []
        for n, contour in enumerate(contours):
            xint = (self.X[0, 0] + contour[:, 1] * self.dl) * (self.X[0, 1] > self.X[0, 0]) + (self.X[0, 0] - contour[:, 1] * self.dl) * (self.X[0, 1] < self.X[0, 0])
            zint = (self.Z[0, 0] + contour[:, 0] * self.dl) * (self.Z[1, 0] > self.Z[0, 0]) + (self.Z[0, 0] - contour[:, 0] * self.dl) * (self.Z[1, 0] < self.Z[0, 0])
            x_int_all.append(xint)
            z_int_all.append(zint)

        flat_xint = []
        flat_zint = []
        for sublist_x, sublist_z in zip(x_int_all, z_int_all):
            for item_x, item_z in zip(sublist_x, sublist_z):
                flat_xint.append(item_x)
                flat_zint.append(item_z)

        x_int_all_array = np.array(flat_xint)
        z_int_all_array = np.array(flat_zint)

        self.ax.plot(x_int_all_array, z_int_all_array, '.k', label='Laplacian edges')
        self.ax.legend()





